import json
import multiprocessing as mp
import os
import shutil
import subprocess
import sys
import tempfile
import time
import numbers
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from typing import Union, List, Dict, Any, Tuple, Optional, Callable
from pprint import pprint

import numpy as np
import pandas as pd
from ebcpy.simulationapi import SimulationAPI, SimulationSetup, SimulationSetupClass
from ebcpy import TimeSeriesData

from pydantic import Field, field_validator


class EPlusSimulationSetup(SimulationSetup):
    """
    Class defining the simulation setup for EnergyPlus simulations.

    Inherits from SimulationSetup base class and adds EnergyPlus specific settings.

    Attributes:
        sim_intervals: List of simulation intervals in format "YYYY-MM-DD to YYYY-MM-DD"
        evaluation_intervals: List of evaluation intervals in format "YYYY-MM-DD to YYYY-MM-DD"
        dataframe_interval: Output interval setting ("Hourly" or "Timestep")
        n_steps_per_hour: Number of simulation timesteps per hour
        _default_solver: Default solver setting
        _allowed_solvers: List of allowed solver options
    """
    sim_intervals: List[Union[str, Tuple[datetime, datetime]]] = Field(
        default=['2024-01-01 to 2024-01-02'],
        description='One or many simulation intervals in format "YYYY-MM-DD to YYYY-MM-DD"',
        title='sim_intervals'
    )

    evaluation_intervals: List[Union[str, Tuple[datetime, datetime]]] = Field(
        default=['2024-01-01 to 2024-01-02'],
        description='One or many evaluation intervals in format "YYYY-MM-DD to YYYY-MM-DD"',
        title='evaluation_intervals'
    )

    dataframe_interval: str = Field(
        default='Timestep',
        description='Interval of the resulting dataframe ("Hourly" or "Timestep")',
        title='dataframe_interval'
    )

    n_steps_per_hour: int = Field(
        default=6,
        description='Number of timesteps per hour',
        title='n_steps_per_hour'
    )

    output_interval: float = Field(
        default=None,  # Set to None as default
        description="The step size of the simulation (automatically set from n_steps_per_hour and/or dateframe_interval)",
        title="output_interval"
    )

    def model_post_init(self, *args, **kwargs):
        """
        Post initialization hook to set output_interval based on n_steps_per_hour.
        This runs after all validators.
        """
        super().model_post_init(*args, **kwargs)
        if self.dataframe_interval == 'Hourly':
            self.output_interval = 3600
        elif self.dataframe_interval == 'Timestep':
            self.output_interval = 3600 / self.n_steps_per_hour

    class Config:
        validate_assignment = True  # Ensures validation happens on attribute assignment

    _default_solver: str = "EPlus"
    _allowed_solvers: List[str] = ["EPlus"]

    @field_validator('sim_intervals', 'evaluation_intervals')
    @classmethod
    def check_intervals(
            cls,
            intervals: List[Union[str, Tuple[datetime, datetime]]]) -> List[Tuple[datetime, datetime]]:
        """
        Validate and convert interval strings or datetime tuples.

        Args:
            intervals: List of either:
                - interval strings in format "YYYY-MM-DD to YYYY-MM-DD"
                - tuples of (start_datetime, stop_datetime)

        Returns:
            List of (start_datetime, stop_datetime) tuples

        Raises:
            ValueError: If intervals are invalid or not chronologically ordered
        """

        if intervals is None:
            return intervals

        _intervals = []
        for interval in intervals:
            # If already a datetime tuple, validate and add directly
            if isinstance(interval, tuple):
                start, stop = interval
                if not (isinstance(start, datetime) and isinstance(stop, datetime)):
                    raise ValueError(
                        'Tuple intervals must contain datetime objects')
            # If string, parse to datetime tuple
            elif isinstance(interval, str):
                start, stop = interval.split(' to ')
                start = datetime.strptime(start, "%Y-%m-%d")
                stop = datetime.strptime(stop, "%Y-%m-%d")
            else:
                raise ValueError(
                    'Intervals must be either strings or datetime tuples')

            if stop <= start:
                raise ValueError('Start must be before stop in intervals')

            _intervals.append((start, stop))

        # Check chronological order
        for n in range(len(_intervals) - 1):
            if _intervals[n][1] > _intervals[n + 1][0]:
                raise ValueError(
                    'Intervals must be increasing in time, meaning the first '
                    'interval is the "oldest" interval')

        return _intervals

    @field_validator('dataframe_interval')
    @classmethod
    def check_dataframe_interval(cls, dataframe_interval: str) -> str:
        """
        Validate the dataframe interval setting.

        Args:
            dataframe_interval: Interval setting ("Hourly" or "Timestep")

        Returns:
            Validated interval setting

        Raises:
            NameError: If interval setting is invalid
        """
        if dataframe_interval not in ['Hourly', 'Timestep']:
            raise NameError(
                'Dataframe Interval must be either "Hourly" or "Timestep"')
        return dataframe_interval


class EnergyPlusAPI(SimulationAPI):
    """
    Class for running EnergyPlus simulations via command line interface.

    Inherits from SimulationAPI base class and implements EnergyPlus specific functionality.
    """

    _sim_setup_class: SimulationSetupClass = EPlusSimulationSetup

    def __init__(self,
                 cd: Path,
                 override_cd: bool,
                 model_path: Path,
                 epw_filepath: Path,
                 energyplus_install_dir: Path,
                 energyplus_version: str,
                 force_same: Optional[Dict] = None,
                 after_sim_functions: Optional[List[Callable]] = None,
                 **kwargs):
        """
        Initialize EnergyPlusAPI instance.

        Args:
            cd: Working directory path
            override_cd: Whether to override existing working directory
            model_path: Path to .idf model file
            epw_filepath: Path to weather file
            energyplus_install_dir: EnergyPlus installation directory
            energyplus_version: EnergyPlus version string
            force_same: Dictionary of parameters to force equal values
            after_sim_functions: List of post-processing functions
            **kwargs: Additional arguments passed to parent class

        Raises:
            ValueError: If working directory exists and override_cd is False
            FileNotFoundError: If model file not found
            TypeError: If model file not .idf format
        """

        if cd.exists() and not override_cd:
            raise ValueError('directory already exists')

        self.energyplus_install_dir = energyplus_install_dir
        assert self.energyplus_install_dir.exists()

        self.epw_filepath = epw_filepath
        assert self.epw_filepath.exists()
        assert self.epw_filepath.suffix == ".epw"

        if not model_path.exists():
            raise FileNotFoundError(model_path)
        if not model_path.suffix == ".idf":
            raise TypeError("Not a valid .idf file")

        self._set_output_timestep = {}
        self.use_for_calibration = False
        self.force_same = force_same
        self.after_sim_functions = after_sim_functions
        self.energyplus_version = energyplus_version
        self.ep_json: Dict[str, Any] = {}

        n_cpu = kwargs.pop("n_cpu", 1)

        super().__init__(
            working_directory=cd,
            model_name=str(model_path),
            n_cpu=n_cpu
        )
        # will be set via the intervals in the simulation setup
        self._start_date = None
        self._stop_date = None

    def _add_sim_setup_parameters(self,
                                  parameters: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Add simulation setup parameters to the parameter dictionary.

        Args:
            parameters: Dictionary of simulation parameters

        Returns:
            Updated parameter dictionary including simulation setup parameters
        """

        n_timesteps_per_hour = self.sim_setup.n_steps_per_hour

        sim_setup = {
            'RunPeriod/Run Period 1/begin_day_of_month': self._start_date.day,
            'RunPeriod/Run Period 1/begin_month': self._start_date.month,
            'RunPeriod/Run Period 1/begin_year': self._start_date.year,
            'RunPeriod/Run Period 1/end_day_of_month': self._stop_date.day,
            'RunPeriod/Run Period 1/end_month': self._stop_date.month,
            'RunPeriod/Run Period 1/end_year': self._stop_date.year,
            'Timestep/Timestep 1/number_of_timesteps_per_hour': n_timesteps_per_hour,
            'Version/Version 1/version_identifier': self.energyplus_version
        }

        if parameters is None:
            return sim_setup

        parameters.update(sim_setup)
        return parameters

    def _get_correct_datetime(self,
                              date_str: str) -> datetime:
        """
        Convert date string to datetime object, handling special case for hour 24.

        Args:
            date_str: Date string to convert

        Returns:
            Converted datetime object
        """

        if date_str[-8:-6] != '24':
            try:
                return pd.to_datetime(date_str, format=" %m/%d  %H:%M:%S")
            except ValueError:
                return pd.to_datetime(date_str, format="%m/%d  %H:%M:%S")

        date_str = date_str.replace(' 24:', ' 00:')
        try:
            return pd.to_datetime(date_str, format=" %m/%d  %H:%M:%S") + \
                timedelta(days=1)
        except ValueError:
            return pd.to_datetime(date_str, format="%m/%d  %H:%M:%S") + \
                timedelta(days=1)

    def _load_csv(self,
                  csv_path: Path) -> pd.DataFrame:
        """
        Load and process simulation results from CSV file.

        Args:
            csv_path: Path to CSV file

        Returns:
            Processed DataFrame with simulation results

        Raises:
            AssertionError: If timedelta consistency check fails
        """

        df = pd.read_csv(csv_path)
        df['Date/Time'] = df['Date/Time'].astype(str)
        df['Date/Time'] = df['Date/Time'].apply(self._get_correct_datetime)

        timedeltas = list(set((i - pd.to_timedelta(i.days, unit='d')).total_seconds()
                              for i in df['Date/Time'].diff()[1:]))

        assert len(timedeltas) == 1
        start_date = self._start_date
        stop_date = self._stop_date

        time_index = pd.date_range(start=start_date,
                                   end=pd.Timestamp(
                                       stop_date) + timedelta(days=1),
                                   freq=f'{timedeltas[0]}s')[1:]

        assert df.shape[0] >= len(time_index)

        df = df.iloc[-len(time_index):].copy()
        df['Date/Time'] = time_index
        df = df.set_index('Date/Time', drop=True)
        return df

    def _single_simulation(self,
                           kwargs: Dict[str, Any]) -> Optional[TimeSeriesData]:
        queue = kwargs.pop('queue', None)
        is_mp = kwargs.pop('is_mp', False)

        _dfs = []

        result_file_name = kwargs.pop("result_file_name", "resultFile")
        result_file_suffix = kwargs.pop("result_file_suffix", "csv")
        parquet_engine = kwargs.pop('parquet_engine', 'pyarrow')
        savepath = kwargs.pop("savepath", None)

        return_option = kwargs.pop("return_option", "time_series")
        if return_option not in ['time_series', 'savepath']:
            raise TypeError(
                f'Return option {return_option} not yet implemented')

        for start, stop in self.sim_setup.sim_intervals:
            self._start_date = start
            self._stop_date = stop

            if is_mp:
                _df = self._single_simulation_solo_mp(kwargs.copy())
            else:
                _df = self._single_simulation_solo(kwargs.copy())
                
            if _df is None:
                if queue is not None:
                    queue.put(None)
                return None
            _dfs.append(_df)

        df = pd.concat(_dfs, axis=0)
        if self.sim_setup.evaluation_intervals is not None:
            df = self._adjust_for_evaluation(df)

        if self.use_for_calibration:
            df = df.reset_index()
            df.index *= self.sim_setup.output_interval
        if queue is not None:
            queue.put(df)

        if return_option == "time_series":
            return df

        if return_option == "savepath":
            if savepath is None:
                savepath = self.working_directory

            os.makedirs(savepath, exist_ok=True)
            filepath = os.path.join(
                savepath, f"{result_file_name}.{result_file_suffix}")
            TimeSeriesData(df).droplevel(1, axis=1).drop(columns=['Date/Time']).save(
                filepath=filepath,
                key="simulation",
                engine=parquet_engine
            )

            return filepath
        return None
    
    def _change_forced_parameters(self,
                                  parameters: Dict[str, Any]) -> Dict[str, Any]:
        """
        Update parameters according to forced parameter relationships.

        Args:
            parameters: Dictionary of simulation parameters

        Returns:
            Updated parameter dictionary with forced relationships applied

        Raises:
            ValueError: If force_from is not a tuple or string
            NameError: If force_from[1] value is not "min" for tuple inputs
        """
        if self.force_same is None:
            return parameters
        
        if parameters is None or len(parameters) == 0:
            return parameters
        
        for force_to, force_from in self.force_same.items():
            if isinstance(force_from, str):
                parameters[force_to] = parameters[force_from]
                continue
            if isinstance(force_from, tuple) and not isinstance(force_from[1], str):
                parameters[force_to] = parameters[force_from[0]] + \
                    force_from[1]
                continue
            
            elif not isinstance(force_from[0], tuple):
                raise ValueError('force_from must be a tuple or a string')
            
            #force_from is a tuple
            if force_from[1] == 'min':
                parameters[force_to] = min([parameters[i] for i in force_from[0]])
            else:
                raise NameError('force_from[1] must be "min"')
        
        return parameters
        
            
    def change_model_dict_with_parameters(
            self,
            model_dict: Dict[str, Any],
            parameters: Dict[str, Any]) -> Dict[str, Any]:
        """
        Update model dictionary with new parameter values.

        Args:
            model_dict: Original model dictionary
            parameters: Dictionary of parameters to update

        Returns:
            Updated model dictionary

        Raises:
            KeyError: If parameter path not found in model dictionary
        """

        def set_value_in_nested_dict(dictionary: dict, 
                             param: str, 
                             value: numbers.Number):
            """
            Set the value of a parameter in a nested dictionary.
            
            Args:
                dictionary (dict): The nested dictionary.
                param (str): The parameter to set.
                value (Any): The value to set the parameter to.
            """
            if not isinstance(value, (numbers.Number, str)):
                raise ValueError(f"Value {value} is not a number or a string")
            
            keys = param.split('/')
            for key in keys[:-1]:
                if "_list" in key:
                    key = int(key[5:])
                dictionary = dictionary[key]
            
            if keys[-1] in ['begin_year', 'end_year'] and keys[-1] not in dictionary:
                if not self.use_for_calibration:
                    print(
                        f"{keys[-1].upper()} NOT INCLUDED IN THE MODEL DESCRIPTION. STILL SIMULATING")
                return
            current_value = dictionary[keys[-1]]
            # check whether the current value is a number or a string
            if not isinstance(current_value, (numbers.Number, str)):
                raise ValueError(f"Current value you are trying to set: {current_value} is not a number or a string")
            dictionary[keys[-1]] = value
        # Change parameter values inside dataframe according to parameters
        for name, value in parameters.items():
            set_value_in_nested_dict(model_dict, name, value)

        return model_dict

    def _single_simulation_solo(self,
                                kwargs: Dict[str, Any]) -> Optional[TimeSeriesData]:
        """
        Run a single simulation in solo mode.

        Args:
            kwargs: Dictionary containing simulation parameters and options

        Returns:
            TimeSeriesData object containing simulation results or None if simulation fails and not fail_on_error

        Raises:
            TypeError: If unsupported return option specified
            ValueError: If inputs provided (not supported)
            Exception: If simulation fails and fail_on_error is True
        """

        parameters = kwargs.pop("parameters", None)
        # Must happen here, because here, parameters might still by empty
        parameters = self._change_forced_parameters(parameters)

        parameters = self._add_sim_setup_parameters(parameters)

        for key in self._set_output_timestep:
            self._set_output_timestep[key] = self.sim_setup.dataframe_interval
        parameters.update(self._set_output_timestep)

        inputs = kwargs.get("inputs", None)
        if inputs is not None:
            raise ValueError("Inputs not supported")

        fail_on_error = kwargs.get("fail_on_error", True)
        ep_json_new = self.ep_json.copy()        
        ep_json_new = self.change_model_dict_with_parameters(ep_json_new,
                                                             parameters)

        _temp_generated_inputs = Path(tempfile.mkdtemp(dir=self.cd))
        generated_input_file = _temp_generated_inputs / 'input.epJSON'

        with open(generated_input_file, 'w') as f:
            json.dump(ep_json_new, f, indent=4)

        _temp_output_directory = Path(tempfile.mkdtemp(dir=self.cd))
        csv_path = _temp_output_directory / 'eplusout.csv'

        cl_st = (f'"{self.energyplus_install_dir}\\EnergyPlus" '
                 + '--readvars '  # included to create a .csv file of the results
                 + '--expandobjects '
                 + f'--output-directory {_temp_output_directory} '
                 + f'--weather "{self.epw_filepath}" '
                 + f'"{generated_input_file}"'
                 )

        # self.logger.debug(cl_st)
        result = subprocess.run(cl_st, capture_output=True)
        success = result.returncode == 0
        if not success:
            if not self.use_for_calibration:
                self.logger.error("Simulation failed!")
                self.logger.error('---ARGS---\n%s', result.args)
                self.logger.error('---STDOUT---\n%s', result.stdout.decode())
                self.logger.error('---STDERR---\n%s', result.stderr.decode())
                error_file = _temp_output_directory / 'eplusout.err'
                if error_file.exists():
                    with open(error_file, 'r') as f:
                        self.logger.error(f.read())
            shutil.rmtree(_temp_output_directory)
            shutil.rmtree(_temp_generated_inputs)
            if fail_on_error:
                raise Exception("Simulation failed")
            return None  # Return nothing

        df = self._load_csv(csv_path)
        shutil.rmtree(_temp_output_directory)
        shutil.rmtree(_temp_generated_inputs)

        if self.after_sim_functions is not None:
            for func in self.after_sim_functions:
                df = func(df)

        return TimeSeriesData(df)

    def _single_simulation_solo_mp(self,
                                   kwargs: Dict[str, Any]) -> Optional[TimeSeriesData]:
        """
        This looks stupid, but the way the multi processing is implemented,
        an error in the simulation will just freeze the progress, because
        the queue is expecting a value, which never comes.

        Returning None is the default behaviour for wrong simulations
        and returning None to the queue will result in the correct handling
        afterwards.
        """
        try:
            return self._single_simulation_solo(kwargs)
        except:
            return None

    def _adjust_for_evaluation(self,
                               df: pd.DataFrame) -> pd.DataFrame:
        """
        Adjust DataFrame to only include specified evaluation intervals.

        Args:
            df: Input DataFrame

        Returns:
            DataFrame filtered to evaluation intervals
        """

        _df_list = []
        for start, stop in self.sim_setup.evaluation_intervals:
            _df = df.loc[(df.index > start) &
                         (df.index <= stop + pd.Timedelta(days=1))].copy()
            _df_list.append(_df)

        return pd.concat(_df_list, axis=0)

    def _single_close(self, **kwargs) -> None:
        """
        Close simulation (placeholder method).
        """
        self.logger.warning(
            "Close functionality does nothing yet. To be implemented...")

    def simulate(self,
                 parameters: Optional[Union[Dict[str, Any],
                                            List[Dict[str, Any]]]] = None,
                 return_option: str = "time_series",
                 **kwargs) -> Union[TimeSeriesData, List[TimeSeriesData], None]:
        """
        Simulate the EnergyPlus model with given parameters.

        Args:
            parameters: Dictionary or list of dictionaries containing simulation parameters
            return_option: How to handle simulation results ("time_series", "last_point", "savepath")
            **kwargs: Additional simulation options

        Returns:
            Simulation results as TimeSeriesData object(s) or None

        Raises:
            TypeError: If saving multiple parameter sets to same path
            ValueError: If multiprocessing parameter lengths mismatch
        """
        # Convert inputs to equally sized objects of lists:
        if parameters is None:
            parameters = [{}]
        if isinstance(parameters, dict):
            parameters = [parameters]
        new_kwargs = {}
        kwargs["return_option"] = return_option  # Update with arg
        # Handle special case for saving files:
        if return_option == "savepath" and len(parameters) > 1:
            savepath = kwargs.get("savepath", [])
            if isinstance(savepath, (str, os.PathLike)):
                savepath = [savepath] * len(parameters)
            result_file_name = kwargs.get("result_file_name", [])
            if (len(set(savepath)) != len(parameters) and
                    len(set(result_file_name)) != len(parameters)):
                raise TypeError(
                    "Simulating multiple parameter set's on "
                    "the same savepath will overwrite old "
                    "results or even cause errors. "
                    "Specify a result_file_name or savepath for each "
                    "parameter combination"
                )
        for key, value in kwargs.items():
            if isinstance(value, list):
                if len(value) != len(parameters):
                    raise ValueError(f"Mismatch in multiprocessing of "
                                     f"given parameters ({len(parameters)}) "
                                     f"and given {key} ({len(value)})")
                new_kwargs[key] = value
            else:
                new_kwargs[key] = [value] * len(parameters)
        kwargs = []
        for _idx, _parameters in enumerate(parameters):
            kwargs.append(
                {"parameters": _parameters,
                 **{key: value[_idx] for key, value in new_kwargs.items()}
                 }
            )
        # Decide between mp and single core
        t_sim_start = time.time()
        if self.use_mp:
            self._n_sim_counter = 0
            self._n_sim_total = len(kwargs)
            n_parallel_runs = int(np.ceil(self._n_sim_total / self.n_cpu))

            self.logger.info(f"Starting {self._n_sim_total} simulations on {self.n_cpu} cores in "
                             f"{n_parallel_runs} parallel runs")

            results = []
            for n_run in range(n_parallel_runs):
                processes = []
                queue = mp.Queue()
                for _ in range(self.n_cpu):
                    if self._n_sim_counter >= len(kwargs):
                        queue.put('Continue')
                        continue

                    _kwargs = kwargs[self._n_sim_counter]
                    self._n_sim_counter += 1

                    _kwargs['queue'] = queue
                    _kwargs['is_mp'] = True
                    p = mp.Process(
                        target=self._single_simulation, args=(_kwargs,))
                    processes.append(p)
                    p.start()

                for _ in range(self.n_cpu):
                    result = queue.get()
                    if isinstance(result, str) and result == 'Continue':
                        continue
                    results.append(result)

                    if self._n_sim_counter == 1 and return_option == 'savepath':
                        self._check_disk_space(result)

                for process in processes:
                    process.join()

                if self._n_sim_counter == self.n_cpu:
                    t1 = time.time()
                if self._n_sim_counter > self.n_cpu:
                    self._remaining_time(t1)

            sys.stderr.write("\r")
        else:
            results = [self._single_simulation(kwargs={
                "parameters": _single_kwargs["parameters"],
                "return_option": _single_kwargs["return_option"],
                **_single_kwargs
            }) for _single_kwargs in kwargs]
        self.logger.info(f"Finished {len(parameters)} simulations on {self.n_cpu} processes in "
                         f"{timedelta(seconds=int(time.time() - t_sim_start))}")
        if len(results) == 1:
            return results[0]
        return results

    def _update_model(self) -> None:
        """
        Update the model by converting IDF to epJSON format and extracting model information.

        This method:
        1. Creates a temporary directory for conversion
        2. Runs EnergyPlus in convert-only mode to create epJSON
        3. Loads the epJSON file and stores it in self.ep_json
        4. Saves the model description
        5. Updates output variable settings

        Raises:
            Exception: If the EnergyPlus conversion fails

        Note:
            Currently, the extraction of model variables (outputs, parameters, etc.) 
            is not implemented.
        """
        output_directory = Path(tempfile.mkdtemp(dir=self.cd))

        cl_st = (f'"{self.energyplus_install_dir}\\EnergyPlus" '
                 + '--convert-only '  # included to create a .csv file of the results
                 + '--expandobjects '
                 + f'--output-directory "{output_directory}" '
                 + f'--weather "{self.epw_filepath}" '
                 + f'"{self.model_name}"'
                 )
        # self.logger.debug(cl_st)

        result = subprocess.run(
            cl_st,
            capture_output=True
        )
        if result.returncode != 0:
            self.logger.error('---ARGS---\n%s', result.args)
            self.logger.error('---STDERR---\n%s', result.stderr.decode())
            self.logger.error('---STDOUT---\n%s', result.stdout.decode())
            shutil.rmtree(output_directory)
            raise Exception("Simulation for model update failed!")

        ep_json_path = output_directory.joinpath(
            Path(self.model_name).stem + ".epJSON")
        with open(ep_json_path, "r") as f:
            self.ep_json = json.load(f)

        with open(Path(self.cd) / 'model_description.json', 'w') as f:
            json.dump(self.ep_json, f, indent=4)

        # Value is set in the simulation since the sim setup is not initialized yet
        for key in self.ep_json['Output:Variable']:
            self._set_output_timestep[f'Output:Variable/{key}/reporting_frequency'] = None

        shutil.rmtree(output_directory)
        self.logger.warning(
            "Extracting model variables (outputs, parameters, etc.) does not yet work. "
            "To be implemented...")

    def convert_json_to_idf(self,
                            json_model_path: Union[str, Path]) -> None:
        """
        Convert an epJSON model file back to IDF format.

        This method:
        1. Creates a temporary directory for the conversion
        2. Loads and saves the JSON model
        3. Uses EnergyPlus to convert it to IDF format
        4. Moves the resulting IDF file to the working directory
        5. Cleans up temporary files

        Args:
            json_model_path: Path to the epJSON model file to convert

        Note:
            The converted IDF file will be named 'calibrated_model.idf' and placed
            in the working directory.
        """

        output_directory = Path(self.cd) / 'converted_model'
        output_directory.mkdir()

        with open(json_model_path, 'r') as f:
            calibrated_model = json.load(f)

        json_model_path = output_directory / 'calibrated_model.epJSON'
        with open(json_model_path, 'w') as f:
            json.dump(calibrated_model, f, indent=4)

        cl_st = (f'"{self.energyplus_install_dir}\\EnergyPlus" '
                 + '--convert-only '  # included to create a .csv file of the results
                 + '--expandobjects '
                 + f'--output-directory "{output_directory}" '
                 + f'--weather "{self.epw_filepath}" '
                 + f'"{json_model_path}"'
                 )

        subprocess.run(
            cl_st,
            capture_output=True
        )

        idf_file = Path(self.cd) / 'converted_model' / 'calibrated_model.idf'
        new_path = Path(self.cd) / 'calibrated_model.idf'

        idf_file.replace(new_path)

        shutil.rmtree(output_directory)
