from pathlib import Path
from typing import List

import pandas as pd
from aixcalibuha import CalibrationClass, SobolAnalyzer, plotting
from data.calibration_config import (ENERGYPLUS_INSTALL_DIR,
                                     ENERGYPLUS_VERSION, GOAL_VAR_CONVERT_DICT,
                                     GOALS, MODEL_PATH, N_CPU,
                                     SIMULATION_SETUP, TUNER_PARAMS,
                                     WEATHER_PATH)

from energyplus_calibrator.calibration_class import EnergyPlusCalibrator
from energyplus_calibrator.energy_plus_api import EnergyPlusAPI


def main():
    measurement_data = pd.read_csv(
        Path(__file__).parents[0] / 'data' / 'measurement_data.csv')
    measurement_data.set_index('Date/Time', inplace=True)
    measurement_data.index = pd.to_datetime(measurement_data.index)

    CD = Path(__file__).parents[1] / 'results' / 'e3_result'

    calibrator = EnergyPlusCalibrator(cd=CD,
                                      model_path=MODEL_PATH,
                                      epw_file_path=WEATHER_PATH,
                                      energyplus_install_dir=ENERGYPLUS_INSTALL_DIR,
                                      energyplus_version=ENERGYPLUS_VERSION,
                                      simulation_setup=SIMULATION_SETUP,
                                      tuner_paras=TUNER_PARAMS,
                                      goals=GOALS,
                                      real_data_dataframe=measurement_data,
                                      goal_var_convert_dict=GOAL_VAR_CONVERT_DICT,
                                      n_cpu=N_CPU)

    run_aixcalibuha_example(working_directory=CD,
                            sim_api=calibrator.energy_plus_api,
                            calibration_class=calibrator.calibration_class,
                            n_cpu=N_CPU)


def run_aixcalibuha_example(working_directory: Path,
                            sim_api: EnergyPlusAPI,
                            calibration_class: List[CalibrationClass],
                            n_cpu: int = 1):
    # This is a copy paste from here: https://github.com/RWTH-EBC/AixCaliBuHA/blob/master/examples/e3_1_verbose_sensitivity_analysis_example.py adjusted to work with the energyplus implementation
    # ## Example of Sobol method
    # First, we instantiate the Analyzer with the `sim_api` and the number of samples.
    # For this example we will use a small sample size to reduce the time needed,
    # which will lead to inaccurate results.
    # In the comparison of the different methods, we will discuss the required sample size.
    # We will also define a working directory were all results of the analysis will be stored.
    # Additionally, we can choose if the samples and corresponding simulation files will be saved.
    # These files can later be loaded and used for analysis of different calibration classes
    # without performing new simulations. The simulations during the sensitivity analysis are the
    # main computational time factor.
    sen_analyzer = SobolAnalyzer(
        sim_api=sim_api,
        num_samples=8,
        calc_second_order=True,
        working_directory=working_directory,
        save_files=True,
        savepath_sim=working_directory.joinpath('files'),
        suffix_files='csv'
    )

    # Now, we run the sensitivity analysis with the verbose option.
    # With that, we not only get the results for combined target values,
    # but we also get the results for every target value alone.
    # Because we defined the first calibration class global
    # and every calibration class has the same
    # tuner-parameters, we can use the option `use_first_sim`,
    # where only the first class will be simulated and the simulation files saved.
    # These simulations will be used for all other classes.

    # For that the simulations are loaded for each class
    # and the statistical measure is then evaluated for the
    # relevant time intervals of each class.

    # When we load simulation files we can use multiprocessing for loading the simulation
    # and evaluating the statistical measure for each class in their time intervals.
    # This multiprocessing option is especially
    # useful for large models and large simulation data,
    # because only one simulation at a time is stored
    # in memory for each process.
    # This can prevent possible memory errors.

    # We disable the automatic plot option here,
    # but we save all results. Later we can use the plot function
    # of the plotting module to plot the results.

    #!! Set load_sim_files to True, if you already did the simulation and only want to load
    # the results and analyse from there
    result, classes = sen_analyzer.run(calibration_classes=calibration_class,
                                       verbose=True,
                                       use_first_sim=True,
                                       plot_result=False,
                                       save_results=True,
                                       n_cpu=N_CPU,
                                       scale=False,
                                       load_sim_files=False)

    # After running the sensitivity analysis you can see
    # that the working directory was created and the result
    # files were saved here. First, the folder "files" was
    # created as the result of `save_files=True`. In there, the
    # simulation files are stored in an own folder which
    # name contains the name of the used calibration class.
    # Additionally, the corresponding samples are stored.
    # The simulations are coupled with their samples.
    # As one results of the analysis, the statistical
    # measure and corresponding sample are saved for each class.
    # This information could be used for surrogate-based
    # calibration, which is currently not implemented in `AixCaliBuHA`.
    # The main results of the sensitivity analysis are the sensitivity
    # measures stored in "SobolAnalyzer_results.csv"
    # and "SobolAnalyzer_results_second_order.csv".
    # These are also returned from the `run()` function as a tuple of two
    # dataframes. This is specific to the sobol method, all other
    # methods only return one dataframe, which is similar to the first return value
    # of the sobol method with possibly other analysis variables.
    # Let´s take a look at these results.

    # The first result has as columns the tuner-parameters
    # and a multi level index with three levels.
    # The first level defines the calibration class.
    # The second level defines the Goals (target values). The index `all`
    # is for the result of the combined target values in the goals.
    # The last level defines the result of the
    # sensitivity measure for each class and goal.
    # These analysis variables are specific for each method.
    # For their exact meaning I refer to the documentation of the SALib or the literature.
    # In this example you get a short overview in the comparison later.
    print("Result of the sensitivity analysis")
    print('First and total order results of sobol method')
    print(result[0].to_string())
    # The second result of the sobol method is for second order sensitive measures.
    # These describe the interaction between two parameters,
    # so this dataframe has a fourth index level "Interaction".
    # In this level, the tuner-parameters are listed again.
    print('Second order results of sobol method')
    print(result[1].to_string())
    # For a better understanding of the results we will now plot them.

    # ## Plotting Sensitivity results
    # We start with the result which were calculated with the small sample size.
    # Let's plot the first and total order results. These results
    # are specific for each single parameter
    # For each calibration class, a figure is created
    # which shows for each goal the first order sensitivity `S1`
    # and the total order sensitivity `ST` combined. For the small
    # sample size the results have large confidence
    # intervals, which show that these results are inaccurate as we
    # noted earlier due to the small sample size.
    plotting.plot_single(result[0])

    # The plotting of second order results is only useful
    # and working for more than 2 parameters.
    # So we only can take a look at them in example A.
    # If you run example B we will skip the plot of the second order results
    # and load some sensitivity results of example A.

    # Let's take a look at the second order results `S2` of example A we just created.
    # This analysis variable shows the interaction of two
    # parameters, so we can plot them as a heatmap.
    # We can see that the parameters have no interaction with
    # themselves what is obvious. Also, we see that the
    # values for p1,p2 and p2,p1 are the same.
    # In the heatmap we can't visualize the confidence intervals,
    # so we will also take a look at the interaction of
    # one specific parameter.
    # For that the `plotting` module has also a function
    # which looks simular to the `S1` and `ST` plots.
    # Here we see again large confidence intervals,
    # so now we will load results which were calculated with
    # a much higher sample number.

    plotting.heatmaps(result[1])
    plotting.plot_single_second_order(
        result[1], 'ZoneInfiltration:DesignFlowRate/IfKJfrdT40ehbMFAOP2OHQ/air_changes_per_hour')


if __name__ == '__main__':
    main()
